import logging
from struct import unpack, pack
import numpy as np
from aug_sfutils import sfdics, str_byt
from aug_sfutils.sfhread import SFH_READ

n_rel = 8

logger = logging.getLogger('aug_sfutils.sfhwrite')
#logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)


for key, val in sfdics.obj_name.items():
    exec('%s=%d' %(val, key))


def object_length(sfobj):

    otyp = sfobj.obj_type
    if otyp not in (Signal, SignalGroup, TimeBase, AreaBase):
        return None
    type_len = sfdics.type_len(sfobj.data_format)

    if otyp == Signal:
        obj_length = type_len * sfobj.index[-1]
    elif otyp == TimeBase:
        obj_length = type_len * sfobj.n_steps
    elif otyp == SignalGroup:
        shape_arr = np.array(sfobj.index[::-1][:sfobj.num_dims])
        obj_length = np.prod(shape_arr) * type_len
    elif otyp == AreaBase:
        shape_arr = np.array([sfobj.size_x + sfobj.size_y + sfobj.size_z, sfobj.n_steps])
        obj_length = np.prod(shape_arr) * type_len

    return obj_length


def param_length(param):

    parmlen = 16 # ParName(=key), unit, dfmt, n_items
    dfmt = param.data_format
    bytlen = param.n_items * sfdics.type_len(dfmt)
    parmlen += 8 * ( (bytlen + 13)//8 )

    return parmlen


def parset_length(pset_d):

    psetlen = 0
    for param in pset_d.values():
        psetlen += param_length(param)

    return psetlen


def par2byt(pname, param):


    dfmt = param.data_format
    n_items = param.n_items

    if dfmt in sfdics.fmt2len.keys(): # char variable
        dlen = sfdics.fmt2len[dfmt]
        bytlen = n_items * dlen
        dj0 = 8 * ( (bytlen + 9)//8 ) 
    elif dfmt in sfdics.fmt2typ.keys(): # number
        sfmt = sfdics.fmt2struc[dfmt]
        type_len = np.dtype(sfmt).itemsize
        val_len = n_items + 2
        bytlen = val_len * type_len
        dj0 = str_byt.next8(bytlen)
    blen = 16 + dj0
    byt = bytearray(blen)
    byt[  :  8] = str_byt.to_byt(pname.ljust(8))
    byt[ 8: 10] = pack('>h', param.physunit)
    byt[10: 12] = pack('>h', dfmt)
    byt[12: 14] = pack('>h', n_items)
    byt[14: 16] = pack('>h', param.status)
    if dfmt in sfdics.fmt2len.keys(): # character variable
        byt[16: 17] = param.dmin
        byt[17: 18] = param.dmax
        param = np.atleast_1d(param)
        for jitem in range(n_items):
            if len(param[jitem]) > 0:
                byt[18 + jitem*dlen: 18 + (jitem+1)*dlen] = param[jitem].ljust(dlen)
    elif dfmt in sfdics.fmt2typ.keys(): # number
        if dfmt == 7: # logical, bug if n_items > 1?
            byt[16: 17] = pack('>%s' %sfmt, param.dmin)
            byt[19: 20] = pack('>%s' %sfmt, param.dmax)
            byt[21: 22] = pack('>%s' %sfmt, param)
        else:
            byt[16         : 16+  type_len] = pack('>%s' %sfmt, param.dmin)
            byt[16+type_len: 16+2*type_len] = pack('>%s' %sfmt, param.dmax)
            param = np.atleast_1d(param)
            for jitem in range(n_items):
                byt[16 + (jitem+2)*type_len: 16 + (jitem+3)*type_len] = pack('>%s' %sfmt, param[jitem])

    return byt


def set_signals(sfo):

# SIGNALS list generated automatically, override input entry
    sigs = type('', (), {})()

    obj_list = []
    for jid in range(len(sfo.sfh)):
        objnam = sfo.getobjname(jid)
        sfobj = sfo.sfh[objnam]
        otyp = sfobj.obj_type
        if otyp in (SignalGroup, Signal, TimeBase, AreaBase): # Include areabases too
            obj_list.append(jid)

    sigs.nitems = len(obj_list)
    sigs.address = len(sfo.sfh)*128
    sigs.length = sigs.nitems*2
    sigs.data_format = 3
    sfmt = sfdics.fmt2struc[sigs.data_format]
    byt_objlist = pack('>%d%s' %(sigs.nitems, sfmt), *obj_list)
    bytlen = len(byt_objlist)
    dj0 = str_byt.next8(bytlen)
    sigs.objlist = bytearray(dj0)
    sigs.objlist[:bytlen] = byt_objlist

    return sigs


def set_length_address(sfo):

# ParSets

    logger.debug('Diag1 %s', sfo.diag)
    len_psets = 0
    for sfobj in sfo.sfh.values(): # sequence not important
        otyp = sfobj.obj_type
        if otyp in (ParamSet, ):
            sfobj.length = parset_length(sfobj.pars)
            len_psets += sfobj.length

# Set lengths and addresses

    sigs = set_signals(sfo)

    addr_diag = sigs.address + sigs.length + len_psets
    par_addr  = sigs.address + sigs.length 
    addr_diag = str_byt.next8(addr_diag)
    par_addr  = str_byt.next8(par_addr)

    addr = addr_diag

    for jid in range(len(sfo.sfh)):

        objnam = sfo.getobjname(jid)
        sfobj = sfo.sfh[objnam]
        otyp = sfobj.obj_type
        addr_in = sfobj.address # For debugging
        len_in  = sfobj.length  # For debugging
        if hasattr(sfobj, 'data_format'):
            type_len = sfdics.type_len(sfobj.data_format)
        else:
            type_len = 0

        if otyp == Diagnostic:
            sfobj.length = sfobj.length # GIT: evaluated at the end of the loop
            sfobj.address = addr
        elif otyp == List:
            if objnam == 'SIGNALS':
                for key, val in sigs.__dict__.items():
                    sfobj.__dict__[key] = val
                addr = addr_diag
        elif otyp in (Device, ParamSet):
            sfobj.address = par_addr
            par_addr += sfobj.length
        elif otyp == Signal:
            sfobj.length = object_length(sfobj)
            sfobj.address = addr
            addr += str_byt.next8(sfobj.length)
        elif otyp == TimeBase:
            sfobj.length = object_length(sfobj)
            sfobj.address = addr
            addr += str_byt.next8(sfobj.length)
        elif otyp == SignalGroup:
            shape_arr = np.array(sfobj.index[::-1][:sfobj.num_dims])
            sfobj.length = object_length(sfobj)
            sfobj.address = addr
            n_block = 2*((sfobj.length + 7)//8)
            addr += n_block*((sfobj.length + n_block-1)//n_block) + shape_arr[1]*2
        elif otyp == AreaBase:
            sfobj.length = object_length(sfobj)
            sfobj.address = addr
            addr += 8*((sfobj.length + 9)//8)
        else:
            continue

        logger.debug('%s (%s)' %(objnam, sfdics.obj_name[otyp]))
        logger.debug('Address in, out: %d %d', addr_in, sfobj.address)
        logger.debug('Length  in, out: %d %d', len_in , sfobj.length )

    len_tot = addr + sfobj.length + addr_diag

    sfobj = sfo.sfh[sfo.diag]
    len_in = sfobj.length
    sfobj.length = len_tot
    logger.debug('%s Length  in, out: %d %d', sfo.diag, len_in , sfobj.length )

    return sfo


def write_sfh(sfo_in, fout):
# Write SFH file

    sfo = set_length_address(sfo_in)
    f = open(fout, 'wb')

    len_sfh = 0
    for jid in range(len(sfo.sfh)):

        objnam = sfo.getobjname(jid)
        sfobj = sfo.sfh[objnam]

# Encode all attributes into byte strings(128)
        sfhbytes = SFH_WRITE(sfobj).bytstr
        f.write(sfhbytes)
        len_sfh += len(sfhbytes)

# Write SIGNALS list

    sigs = sfo.sfh['SIGNALS']
    f.write(sigs.objlist)
    len_sfh += len(sigs.objlist)

# Write content of ParSets
    for jid in range(len(sfo.sfh)):

        objnam = sfo.getobjname(jid)
        sfobj = sfo.sfh[objnam]
        otyp = sfobj.obj_type
        
        if otyp in (ParamSet, ):
            pset2byt = bytearray(parset_length(sfobj.pars))
            j0 = 0
            for pname, param in sfobj.pars.items():
                p2b = par2byt(pname, param)
                j1 = j0 + len(p2b)
                pset2byt[j0: j1] = p2b
                j0 = j1
            f.write(pset2byt)
    f.close()
    logger.info('Stored binary %s' %fout)


class SFH_WRITE:
    """
    Writes a generic SFH object metadata to a byte string
    """

    def __init__(self, sfo):
        """
        Writes the SFH part of an object
        """
        self.sfo = sfo
        objnam   = str_byt.to_byt(sfo.objnam)
        obj_type = pack('>H', sfo.obj_type)
        level    = pack('>H', sfo.level)
        status   = pack('>H', sfo.status)
        errcode  = pack('>h', sfo.errcode)
        rel      = pack('>8H', *sfo.rel)
        address  = pack('>I', sfo.address)
        length   = pack('>I', sfo.length)
        self.val = bytearray(24)
        descr    = str_byt.to_byt(sfo.descr)

        val_func = { \
            1 : self.sfdiag, 2: self.list, 3: self.device, 4: self.parmset, \
            5 : self.mapping, 6: self.sig, 7: self.sig, 8: self.tbase, \
            9 : self.sf_list, 10: self.algorithm, 11: self.update_set, \
            12: self.loctimer, 13: self.abase, 14: self.qualifier, \
            15: self.modobj, 16: self.mapext, 17: self.resource, \
            18: self.addrlen }

        if sfo.obj_type in val_func.keys():
#            logger.debug('%d %s %d', sfo.objid, sfo.objnam, sfo.obj_type)
            val_func[sfo.obj_type]()
        else:
            logger.warning('Object type %d not supported' %sfo.obj_type)

        self.bytstr = bytearray(128)
        self.bytstr[  :  8] = objnam.ljust(8)
        self.bytstr[ 8: 10] = obj_type
        self.bytstr[10: 12] = level
        self.bytstr[12: 14] = status
        self.bytstr[14: 16] = errcode
        self.bytstr[16: 32] = rel
        self.bytstr[32: 36] = address
        self.bytstr[36: 40] = length
        self.bytstr[40: 64] = self.val 
        self.bytstr[64:128] = descr.ljust(64)


    def sfdiag(self):
        """
        Metadata of a DIAG object
        """
        self.val[  :  4] = self.sfo.diag_code
        self.val[ 4:  6] = pack('>H', self.sfo.num_objs)
        self.val[ 6:  8] = pack('>H', self.sfo.diag_type)
        self.val[ 8: 12] = pack('>I', self.sfo.shot_nr)
        self.val[12: 16] = pack('>I', self.sfo.c_time)
        self.val[16: 18] = pack('>H', self.sfo.up_limit)
        self.val[18: 20] = pack('>H', self.sfo.exp)
        self.val[20: 22] = pack('>H', self.sfo.version)
        self.val[22: 24] = pack('>H', self.sfo.s_type)


    def list(self):
        """
        Metadata of a LIST object
        """

        self.val[ : 2] = pack('>H', self.sfo.data_format)
        self.val[2: 4] = pack('>H', self.sfo.nitems)
        self.val[4: 6] = pack('>H', self.sfo.ordering)
        self.val[6: 8] = pack('>H', self.sfo.list_type)


    def device(self):
        """
        Metadata of a DEVICE object
        """

        self.val[  :  2] = pack('>H', self.sfo.data_format)
        self.val[ 2:  4] = pack('>H', self.sfo.acquseq)
        self.val[ 4:  6] = pack('>H', self.sfo.nitems)
        self.val[ 6:  8] = pack('>H', self.sfo.dev_type)
        self.val[ 8: 12] = pack('>I', self.sfo.dev_addr)
        self.val[12: 16] = pack('>I', self.sfo.n_chan)
        self.val[16: 18] = pack('>H', self.sfo.task)
        self.val[18: 20] = pack('>H', self.sfo.dev_num)
        self.val[20: 24] = pack('>I', self.sfo.n_steps)


    def parmset(self):
        """
        Metadata of a ParameterSet object
        """

        self.val[4: 6] = pack('>H', self.sfo.nitems)
        self.val[6: 8] = pack('>H', self.sfo.cal_type)


    def mapping(self):
        """
        Metadata of a MAP object
        """

        self.val[ 4:  6] = pack('>H', self.sfo.nitems)
        self.val[ 6:  8] = pack('>H', self.sfo.map_type)
        self.val[16: 18] = pack('>H', self.sfo.task)


    def sig(self):
        """
        Metadata of a Signal or SignalGroup object
        """
        ind = list(self.sfo.index)[::-1]
        while len(ind) < 4:
            ind.append(1)
        self.sfo.index = ind[::-1]

        self.val[  :  2] = pack('>H' , self.sfo.data_format)
        self.val[ 2:  4] = pack('>H' , self.sfo.physunit)
        self.val[ 4:  6] = pack('>H' , self.sfo.num_dims)
        self.val[ 6:  8] = pack('>h' , self.sfo.stat_ext)
        self.val[ 8: 24] = pack('>4I', *self.sfo.index)


    def tbase(self):
        """
        Metadata of a TIMEBASE object
        """

        self.val[  :  2] = pack('>H', self.sfo.data_format)
        self.val[ 2:  4] = pack('>H', self.sfo.burstcount)
        self.val[ 4:  6] = pack('>H', self.sfo.event)
        self.val[ 6:  8] = pack('>H', self.sfo.tbase_type)
        self.val[ 8: 12] = pack('>I', self.sfo.s_rate)
        self.val[16: 20] = pack('>I', self.sfo.n_pre)
        self.val[20: 24] = pack('>I', self.sfo.n_steps)


    def sf_list(self):
        """
        Metadata of a SF_LIST object
        """

        self.val[2: 4] = pack('>H', self.sfo.nitems)


    def algorithm(self):
        """
        Metadata of an ALGORITHM object
        """

        self.val[ 8: 16] = self.sfo.hostname
        self.val[16: 24] = self.sfo.date


    def update_set(self):
        """
        Metadata of an UPDATE_SET object
        """

        self.val[ 2:  4] = pack('>H', self.sfo.nitems) 
        self.val[ 4:  8] = pack('>i', self.sfo.input_vals) 
        self.val[16: 20] = pack('>I', self.sfo.next_index) 
        self.val[20: 24] = pack('>I', self.sfo.size) 


    def loctimer(self):
        """
        Metadata of an LOCTIMER object
        """

        self.val[  :  2] = pack('>H', self.sfo.data_format)
        self.val[ 2:  4] = pack('>H', self.sfo.resolution)
        self.val[20: 24] = pack('>I', self.sfo.size)


    def abase(self):
        """
        Metadata of an AREABASE object
        """

        self.val[  :  2] = pack('>H', self.sfo.data_format)
        self.val[ 2:  8] = pack('>3H', *self.sfo.physunit)
        self.val[ 8: 12] = pack('>I', self.sfo.size_x)
        self.val[12: 16] = pack('>I', self.sfo.size_y)
        self.val[16: 20] = pack('>I', self.sfo.size_z)
        self.val[20: 24] = pack('>I', self.sfo.n_steps)


    def qualifier(self):
        """
        Metadata of an QUALIFIER object
        """

        self.val[  :  2] = pack('>H', self.sfo.data_format)
        self.val[ 4:  6] = pack('>H', self.sfo.num_dims)
        self.val[ 6:  8] = pack('>H', self.sfo.qsub_typ)
        self.val[ 8: 12] = pack('>I', self.sfo.index_4)
        self.val[12: 16] = pack('>I', self.sfo.index_3)
        self.val[16: 20] = pack('>I', self.sfo.index_2)
        self.val[20: 24] = pack('>I', self.sfo.max_sections)


    def modobj(self):
        """
        Metadata of a MODOBJ object
        """

        self.val[ : 2] = pack('>H', self.sfo.nitems)


    def mapext(self):
        """
        Metadata of a MAPEXT object
        """

        self.val[  :  2] = pack('>H', self.sfo.nitems)
        self.val[ 2:  4] = pack('>H', self.sfo.mapalg)
        self.val[ 4:  8] = pack('>I', self.sfo.tbeg)
        self.val[ 8: 12] = pack('>I', self.sfo.tend)
        self.val[12: 14] = pack('>H', self.sfo.val_0)
        self.val[14: 16] = pack('>H', self.sfo.val_1)
        self.val[16: 20] = pack('>I', self.sfo.val_2)
        self.val[20: 24] = pack('>I', self.sfo.val_3)


    def resource(self):
        """
        Metadata of a RESOURCE object
        """

        self.val[  :  2] = pack('>H', self.sfo.num_cpus)
        self.val[ 2:  4] = pack('>H', self.sfo.first_cpu)


    def addrlen(self):
        """
        Value of ADDRLEN object
        """

        self.val[ : 2] = pack('>H', self.sfo.addrlen)



"""Class for shotfiles writing"""

import os, datetime, warnings, logging
import ctypes as ct
import numpy as np
from aug_sfutils import sfdics, str_byt

logger = logging.getLogger('aug_sfutils.ww')
#logger.setLevel(logging.INFO)
#logger.setLevel(logging.DEBUG)


wwlib = '/afs/ipp-garching.mpg.de/aug/ads/lib64/@sys/libddww8.so.8.1'
if not os.path.isfile(wwlib):
    wwlib = '/afs/ipp-garching.mpg.de/aug/ads/lib64/amd64_sles11/libddww8.so.8.1'
libddww = ct.cdll.LoadLibrary(wwlib)

for key, val in sfdics.obj_name.items():
    exec('%s=%d' %(val, key))
for key, val in sfdics.typ2descr.items():
    if key is not None:
        exec('%s=%d' %(val, key))


def getError(error):

    """ Check if an error/warning occured
    """

    try:
        err = ct.c_int32(error)
    except TypeError:
        err = ct.c_int32(error.value)
    isError = libddww.xxsev_(ct.byref(err)) == 1
    isWarning = libddww.xxwarn_(ct.byref(err)) == 1
    if isError or isWarning:
        id = ct.c_char_p(b'')
        lid = ct.c_uint64(0)
        text = ct.c_char_p(b' '*255)
        ltext = ct.c_uint64(255)
        unit = ct.byref(ct.c_int32(-1))
        ctrl = ct.byref(ct.c_uint32(3))
        libddww.xxerrprt_(unit, text, ct.byref(err), ctrl, id, ltext, lid);
        if isError:
            raise Exception(text.value.decode().strip())
        else:
            warnings.warn(text.value.decode().strip(), RuntimeWarning)


class WW:

    """py-WW for shotfile writing
    Wrapper around libddww8.so
    """

    def Open(self, exp, diag, nshot, edition=-1, mode='new'):
        """Equivalent of wwopen
        Prepares output shotfile.
        """
        self.Close()
        if nshot > 0:
            date = datetime.date.strftime(datetime.datetime.today(),'%y.%m.%d;%H:%M:%S')

            exp  = str_byt.to_byt(exp)
            diag = str_byt.to_byt(diag)
            mode = str_byt.to_byt(mode)
            date = str_byt.to_byt(date)

            self.edition = ct.c_int32(edition)
            error        = ct.c_int(0)
            diaref       = ct.c_int32(0)
            shot         = ct.c_uint32(nshot)
            cexp         = ct.c_char_p(exp )
            cdiag        = ct.c_char_p(diag)
            cmode        = ct.c_char_p(mode)
            cdate        = ct.c_char_p(date)
            _error       = ct.byref(error)
            _nshot       = ct.byref(shot)
            _edition     = ct.byref(self.edition)
            self._diaref = ct.byref(diaref)
            lexpr = ct.c_uint64(len(exp ))
            ldiag = ct.c_uint64(len(diag))
            lmode = ct.c_uint64(len(mode))
            ldate = ct.c_uint64(len(date))

            result = libddww.wwopen_(_error, cexp, cdiag, _nshot, cmode, _edition, \
                                     self._diaref, cdate, lexpr, ldiag, lmode, ldate)
            if getError( error ):
                del self._diaref
                del self.edition
                raise Exception('ww: Error Opening Shotfile %(error)s' % {'error':error.value})
        return hasattr(self, '_diaref')


    def Close(self):
        """Equivalent of wwclose.
        Terminates the file stream and deletes some attributes
        """
        if hasattr(self, '_diaref'):
            error = ct.c_int32(0)
            disp  = b'lock'
            space = b'maxspace'
            ldisp  = ct.c_uint64(len(disp))
            lspace = ct.c_uint64(len(space))
            _error = ct.byref(error)
            _disp  = ct.c_char_p( disp )
            _space = ct.c_char_p( space )

            result = libddww.wwclose_(_error , self._diaref , _disp , _space , ldisp , lspace )
            logger.info('Close: %d, Error: %d, Edition: %d', result, error.value, self.edition.value)
            if getError( error ):
                raise Exception('ww: Error Closing Shotfile')
            del self._diaref
            del self.edition


    def _wwoinfo(self, signame):
        """Wrapping wwoinfo_
        """

        if hasattr(self, '_diaref'):
            signame = str_byt.to_byt(signame)

            error   = ct.c_int32(0)
            name    = ct.c_char_p(signame)
            lname   = ct.c_uint64(len(signame))
            typ     = ct.c_uint32(0)
            format  = ct.c_uint16(0)
            ntval   = ct.c_uint32(0)
            items   = ct.c_uint32(0)
            _items  = ct.byref(items)
            c_indices = (ct.c_uint32 * 4)()
            _error   = ct.byref(error)
            _typ     = ct.byref(typ)
            _ntval   = ct.byref(ntval)
            _format  = ct.byref(format)
            _indices = ct.byref(c_indices)

#            logger.debug(signame)
            result = libddww.wwoinfo_(_error, self._diaref, name, _typ, \
                     _format, _ntval, _items, _indices, lname)

            if getError(error):
                lbl = sfdics.obj_name[typ.value]
                raise Exception('ww: Error wwoinfo %s' %(lbl, signame) )

            return {'otype': typ.value, 'format': format.value, 'leng': ntval.value, \
                    'items': items.value, 'indices': c_indices}


    def GetParameterInfo(self, pset, pnam):
        """Returns information about the parameter 'pnam' of the parameter set 'pset'."""

        if hasattr(self, '_diaref'):
            par_name = str_byt.to_byt(pnam)
            set_name = str_byt.to_byt(pset)
            error   = ct.c_int32(0)
            _error  = ct.byref(error)
            pset    = ct.c_char_p(set_name)
            par     = ct.c_char_p(par_name)
            items   = ct.c_uint32(0)
            _items  = ct.byref(items)
            format  = ct.c_uint16(0)
            _format = ct.byref(format)
            lpar = ct.c_uint64(len(par_name))
            lset = ct.c_uint64(len(set_name))
            result = libddww.dd_prinfo_(_error, self._diaref, pset, par, _items, _format, lset, lpar)
            return {'items': items.value, 'format': format.value}


    def SetObject(self, signame, data, tim_axis=None):
        """Stores 'data'-array of a given object named 'signame' of type 
        Signal, SigGroup or Areabase to the target shotfile
        """

        logger.info('Setting %s', signame)
        if hasattr(self, '_diaref'):
            info  = self._wwoinfo(signame)
            otype = info['otype']
            if otype in (Signal, TimeBase):
                self.SetSignal(signame, data)
            elif otype == SignalGroup:
                self.SetSignalGroup(signame, data)
            elif otype == AreaBase:
                self.SetAreabase(signame, data, tim_axis=tim_axis)


    def SetSignal(self, signame, data, indices=None):
        """Stores 'data'-array of a given Signal or TimeBase named 'signame'
        to the target shotfile
        """

        if hasattr(self, '_diaref'):
            info  = self._wwoinfo(signame)
            dfmt = info['format']
            otype = info['otype']
            if dfmt in sfdics.fmt2len.keys():
                lbuf = sfdics.fmt2len[dfmt]
                lsbuf = ct.c_uint64(lbuf)
                sf_dtype = CHAR
            else:
                lbuf = np.size(data)
                sf_dtype = sfdics.fmt2typ[dfmt]

            signame = str_byt.to_byt(signame)
            lbl = sfdics.obj_name[otype]

            error   = ct.c_int32(0)
            name    = ct.c_char_p(signame )
            lname   = ct.c_uint64(len(signame))
            clbuf   = ct.c_uint32(lbuf)
            stride  = ct.c_uint32(1)
            _error  = ct.byref(error)
            _type   = ct.byref(ct.c_uint32(sf_dtype))
            _lbuf   = ct.byref(clbuf)
            _stride = ct.byref(stride)
            if indices is not None:
                _indices = np.array(indices, dtype=np.int32).ctypes.data_as(ct.POINTER(ct.c_uint32))

            if dfmt in sfdics.fmt2len.keys():
                slen = sfdics.fmt2len[dfmt]
#                stride  = ct.c_uint32(slen)
                stride  = ct.c_uint32(1)
                _stride = ct.byref(stride)
                if type(data) == type('str'):
                    sbuf = data.ljust(slen)
                else: #string array
                    sbuf = ''
                    for entry in data:
                        sbuf += entry.ljust(slen)

                sbuf = str_byt.to_byt(sbuf)
                lbuf = ct.c_uint32(len(sbuf))
                _lbuf = ct.byref(lbuf)
                lsbuf = ct.c_uint64(len(sbuf)) 
                buffer = ct.c_char_p(sbuf)
                if otype == Signal:
                    result = libddww.wwsignal_( \
                             _error, self._diaref, name, _type, _lbuf, buffer, \
                             _stride, lname, lsbuf)
                elif otype == SignalGroup:
                    result = libddww.wwinsert_( \
                             _error, self._diaref, name, _type, _lbuf, buffer, \
                             _stride, _indices, lname, lsbuf)
            else:
                ctyp = sfdics.fmt2ct[dfmt]
                data = np.atleast_1d(data)
                data = np.array(data, dtype=sfdics.fmt2np[dfmt])
                if otype == Signal:
                    result = libddww.wwsignal_( \
                             _error, self._diaref, name, _type, _lbuf, \
                             data.ctypes.data_as(ct.POINTER(ctyp)), \
                             _stride, lname)
                elif otype == TimeBase:
                    result = libddww.wwtbase_( \
                             _error, self._diaref, name, _type, _lbuf, \
                             data.ctypes.data_as(ct.POINTER(ctyp)), \
                             _stride, lname)
                elif otype == SignalGroup:
                    result = libddww.wwinsert_( \
                             _error, self._diaref, name, _type, _lbuf, \
                             data.ctypes.data_as(ct.POINTER(ctyp)), \
                             _stride, _indices, lname)

            if otype in (Signal, TimeBase):
                sign = signame.decode('utf8')
                if getError( error ):
                    raise Exception('ww: Error Writing %s %s' %(lbl, sign) )
                else:
                    logger.info('Written %s %s', lbl, sign)


    def SetSignalGroup(self, signame, data):
        """Stores 'data'-array of a given SignalGroup named 'signame'
        to the target shotfile
        """

        info  = self._wwoinfo(signame)
        dfmt = info['format']
        if dfmt in sfdics.fmt2len.keys():
            for i, dat in enumerate(data):
                dat = dat.ljust(info['leng'])
                indices = [i + 1, 1, 1]
                self.SetSignal(signame, dat, indices=indices)
        else:
            if data.ndim < 2:
                raise Exception('ww: SetSignalGroup array %s has only %d dims' %(signame, data.ndim))
                return False
            if data.ndim == 2:
                for i in range(data.shape[1]):
                    indices = [i + 1, 1, 1]
                    self.SetSignal(signame, data[:, i], indices=indices)
            if data.ndim == 3:
                for i in range(data.shape[1]):
                    for j in range(data.shape[2]):
                        indices = [1 + i, 1 + j, 1]
                        self.SetSignal(signame, data[:, i, j], indices=indices)
        logger.info('Written Signal Group %s', signame)

        return True


    def SetAreabase(self, areaname, data, sizes=None, tim_axis=None):
        """Stores 'data'-array of a given Areabase named 'areaname'
        to the target shotfile
        """

        if hasattr(self, '_diaref'):
            info  = self._wwoinfo(areaname)
            dfmt = info['format']
            if dfmt in sfdics.fmt2len.keys():
                raise Exception('AB cannot have string-like data')
            sf_dtype = sfdics.fmt2typ[dfmt]
            ctyp  = sfdics.fmt2ct[dfmt] # data have to be numerical for AB

            data = np.array(data, dtype=sfdics.fmt2np[dfmt])
            areaname = str_byt.to_byt(areaname)     
            csizes = (ct.c_uint32 * 3)()
            if tim_axis is not None and tim_axis != 0:
                logger.error('tim_axis can be only None or 0, not %d', tim_axis)
                return

            if data.ndim > 2:
                logger.error('AB %s hat %d dims: only a time and a space dependence are allowed, flatten the space data', areaname, data.ndim)
                return

            if tim_axis is None:
                nt = 1
                data = np.atleast_2d(data).T
            else:
                if tim_axis == 0:
                    data = data.T
                nt = data.shape[1]

            if sizes is None:
                csizes[0] = data.shape[0]
            else:
                nx = len(sizes)
                csizes[:nx] = sizes[:nx]


            logger.debug('AB %s, %d, %d, %d, %d, %d', areaname, sf_dtype, nt, *csizes)

            error  = ct.c_int32(0)
            name   = ct.c_char_p(areaname)
            lname  = ct.c_uint64(len(areaname))
            k1     = ct.c_uint32(1) # c_long
            k2     = ct.c_uint32(nt)
            _k1    = ct.byref(k1)
            _k2    = ct.byref(k2)
            _type  = ct.byref(ct.c_uint32(sf_dtype))
            _error = ct.byref(error)
            _sizes = ct.byref(csizes)

            libddww.wwainsert_(_error, self._diaref, name, _k1, _k2, _type, \
                               data.ctypes.data_as(ct.POINTER(ctyp)), \
                               _sizes, lname)

            if getError(error):
                raise Exception('ww.shotfile.SetAreabase: Error writing areabase %s' %areaname)
            else:
                logger.info('Written Areabase %s, sizes %d %d %d', areaname, *csizes[:3])


    def SetParameter(self, set_name, par_name, data):

        """ ww.shotfile().SetParameter( SetName , ParameterName , data )
        types: 0=raw 1=integer 2=float 3=double 4=complex 5=logical 6=character
        The data type is derived from the 00000 shotfile.
        """

        if hasattr(self, '_diaref'):
            info = self.GetParameterInfo(set_name, par_name)
            dfmt = info['format']
            if dfmt in sfdics.fmt2len.keys():
                sf_dtype = CHAR
            else:
                sf_dtype = sfdics.fmt2typ[dfmt]
            set_name = str_byt.to_byt(set_name)
            par_name = str_byt.to_byt(par_name)
            logger.info('PN %s', par_name)
            data = np.atleast_1d(data)
            ndata = len(data)
            nitem = info['items']
            if ndata > nitem:
                logger.error('Too many %d entries for par_nam %s, more than in sfh %d', ndata, par_name, nitem)
                return
            if ndata == 0:
                logger.error('No data for par_nam %s', par_name)
                return
#            if ndata < nitem:
#                data = np.zeros(nitem, dtype=sfdics.typ2np[sf_dtype])
            error   = ct.c_int32(0)
            pset    = ct.c_char_p(set_name)
            par     = ct.c_char_p(par_name)
            lset    = ct.c_uint64(len(set_name))
            lpar    = ct.c_uint64(len(par_name))
            _type   = ct.byref(ct.c_uint32(sf_dtype))
            _error  = ct.byref(error)
            if dfmt in sfdics.fmt2len.keys(): # string or string array
                slen = sfdics.fmt2len[dfmt]
                stride  = ct.c_uint32(1)
                _stride = ct.byref(stride)
                if type(data) == type('str'):
                    sbuf = data.ljust(slen)
                else: #string array
                    sbuf = ''
                    for entry in data:
                        sbuf += entry.ljust(slen)

                sbuf = str_byt.to_byt(sbuf)
                lbuf = ct.c_uint32(len(sbuf))
                _lbuf = ct.byref(lbuf)
                lsbuf = ct.c_uint64(len(sbuf)) 
                buffer = ct.c_char_p(sbuf)
                result = libddww.wwparm_(_error, self._diaref, pset, par, _type, _lbuf, \
                                         buffer, _stride, lset, lpar, lsbuf)
            else: # numerical data
                ctyp  = sfdics.fmt2ct[dfmt]
                datanp = np.array(data, dtype=sfdics.fmt2np[dfmt])
                if ndata < nitem:
                    datanp = np.append(datanp, np.zeros(nitem - ndata))
                stride  = ct.c_uint32(1)
                _stride = ct.byref(stride)
                lbuf = ct.c_uint32(nitem)
                _lbuf = ct.byref(lbuf)
                if nitem == 1 and sf_dtype == LOGICAL:
                    buffer = ctyp(datanp)
                else:
                    buffer = ctyp.from_buffer(datanp)
                _buffer = ct.byref(buffer)
                result = libddww.wwparm_(_error, self._diaref, pset, par, _type, _lbuf, \
                                         _buffer, _stride, lset, lpar)
            if getError( error ):
                raise Exception('ww: Error Writing Parameters into %s -> %s' %(set_name.decode('utf-8'), par_name.decode('utf-8') ) )

            else:
                logger.info('   PN %s', par_name.decode('utf8'))


def write_sf(nshot, data_d, sfhdir, diag, exp='AUGD'):
    """Function for writing a full shotfile from a data dict

    Input:
        nshot    int   Shotnumber
        data_d   dict  Dictionary with all relevant objects;
            Keys are object names - they must match the names in the 00000-SFH
            Values are object arrays or dictionaries for ParSets. The data
            types must be the numpy equivalents of the types declared in the
            00000-SFH.
            The array sizes (dict values) have to be consistent among them
            and with the relations in the 00000-SFH, but they do not need
            to be equal to the ones in the 00000-SFH: if the aren't, 
            run sfh.Mod* ahead
        sfhdir   str  Full path of the 00000-SFH
        diag     str  Diagnostic name
        exp(opt) str  Exp name
    """

    import os

    ww = WW()

    os.chdir(sfhdir)
    logger.info('Exp: %s, diag: %s, nshot: %d', exp, diag, nshot)
    if not ww.Open(exp, diag, nshot):
        logger.error('Problems opening shotfile')
        return

    otype = {}
    for obj, data in data_d.items():
        otype[obj] = ww._wwoinfo(obj)['otype']
        if otype[obj] == TimeBase:
            ww.SetObject(obj, data)

    for obj, data in data_d.items():
        if otype[obj] == AreaBase:
            if data.ndim == 1:
                tim_ax = None
            else:
                tim_ax = 0
            ww.SetObject(obj, data, tim_axis=tim_ax)

    for obj, data in data_d.items():
        if otype[obj] == ParamSet:
            logger.info('Writing PS %s', obj)
            for pn, pdata in data.items():
                ww.SetParameter(obj, pn, pdata)
        if otype[obj] in (Signal, SignalGroup):
            ww.SetObject(obj, data)

    ww.Close()
